//
//  AppDelegate.m
//  DDC Panel - Hackintosh Brightness
//
//  Created by PHPdev32 on 2/19/13.
//  Updated for f-keys and mac-like OSD support by Konstantinos Nks on 07/23/15
//  Licensed under GPLv3, full text at http://www.gnu.org/licenses/gpl-3.0.txt
//

#import "AppDelegate.h"
#import "DDC.h"
#import <IOKit/hidsystem/ev_keymap.h>

@implementation FSApplication

-(void)sendEvent:(NSEvent *)theEvent{
    [super sendEvent:theEvent];
    if (theEvent.type == NSSystemDefined
        && theEvent.subtype == NX_SUBTYPE_AUX_CONTROL_BUTTONS
        && theEvent.data2 == NSUIntegerMax
        && (theEvent.data1 >> 8 & 0xFF) == NX_KEYDOWN
        && [self.delegate respondsToSelector:@selector(mediaKeyDown:)])
        [self.delegate performSelector:@selector(mediaKeyDown:) withObject:@(theEvent.data1 >> 16 & 0xFF)];
}

@end

@implementation DDC {
@private
    NSUInteger _control;
    dispatch_semaphore_t _block;
    CGDirectDisplayID _display;
    NSNumber *_value;
}

@synthesize max = _max;

+(NSString *)EDIDString:(char *)string{
    NSString *temp = [[NSString alloc] initWithBytes:string length:13 encoding:NSASCIIStringEncoding];
    return ([temp rangeOfString:@"\n"].location != NSNotFound)
    ? [[temp componentsSeparatedByString:@"\n"] objectAtIndex:0]
    : temp;
}

-(id)initWithControl:(NSUInteger)control on:(CGDirectDisplayID)display {
    self = [super init];
    if (self) {
        _enabled = true;
        _block = dispatch_semaphore_create(1);
        _display = display;
        _control = control;
    }
    return self;
}
-(NSUInteger)max {
    if (!_value)
        [self value];
    return _max;
}
-(NSUInteger)value{
    if (!_value) {
        dispatch_semaphore_wait(_block, DISPATCH_TIME_FOREVER);
        struct DDCReadCommand command = {_control};
        if (!DDCRead(_display, &command)) {
            muteWithNotice(self, enabled, _enabled = false);
            _value = @0;
        }
        else
            _value = @(command.current_value);
        dispatch_semaphore_signal(_block);
        assignWithNotice(self, max, command.max_value);
        delayWithNotice(self, value, 0);
    }
    return _value.unsignedLongValue;
}
-(void)setValue:(NSUInteger)value{
    if (!_enabled)
        return;
    dispatch_semaphore_wait(_block, DISPATCH_TIME_FOREVER);
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
        struct DDCWriteCommand command = {_control, value};
        if (!DDCWrite(_display, &command)) {
            muteWithNotice(self, enabled, _enabled = false);
        }
        dispatch_semaphore_signal(_block);
    });
    _value = nil;
}

@end

@implementation AppDelegate {
@private
    NSPointerArray *_displayIDs;
}

#pragma mark NSApplicationDelegate
-(void)awakeFromNib {
    //[NSApp setActivationPolicy: NSApplicationActivationPolicyAccessory];
    _displayIDs = [NSPointerArray pointerArrayWithOptions:NSPointerFunctionsOpaqueMemory | NSPointerFunctionsIntegerPersonality];
    [self applicationDidChangeScreenParameters:nil];
    [self addObserver:self forKeyPath:@"displayNumber" options:NSKeyValueObservingOptionInitial context:0];
}

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)sender{
    return true;
}

-(void)applicationDidChangeScreenParameters:(NSNotification *)notification {
    CGDirectDisplayID old = _displayNumber && _displayNumber <= _displayIDs.count
    ? (CGDirectDisplayID)[_displayIDs pointerAtIndex:_displayNumber - 1]
    : 0;
    while (_displayIDs.count)
        [_displayIDs removePointerAtIndex:0];
    for (NSScreen *screen in NSScreen.screens) {
        CGDirectDisplayID new = [[screen.deviceDescription objectForKey:@"NSScreenNumber"] unsignedIntValue];
        [_displayIDs addPointer:(void *)(UInt64)new];
        if (new == old && _displayNumber != _displayIDs.count)
            self.displayNumber = _displayIDs.count;
    }
    _displaysView.numberOfTickMarks = NSScreen.screens.count;
    assignWithNotice(self, allDisplays, _displaysView.numberOfTickMarks);
  
    

    if (!old)
        self.displayNumber = 1;
}

#pragma mark Actions
-(IBAction)refresh:(id)sender {
    dispatch_after(dispatch_time(DISPATCH_TIME_NOW, 2 * NSEC_PER_SEC), dispatch_get_main_queue(), ^{
        [self observeValueForKeyPath:nil ofObject:nil change:nil context:NULL];
    });
}
BOOL checkAccessibility()
{
    NSDictionary* opts = @{(__bridge id)kAXTrustedCheckOptionPrompt: @YES};
    return AXIsProcessTrustedWithOptions((__bridge CFDictionaryRef)opts);
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
   
    if (checkAccessibility()) {
        NSLog(@"Accessibility Enabled");
    }
    else {
        NSLog(@"Accessibility Disabled");
    }

    [NSEvent addGlobalMonitorForEventsMatchingMask:NSKeyDownMask
                                           handler:^(NSEvent *event){
                                               [self mediaKeyDown:@(event.keyCode)];
                                               
                                           }];
}
- (void)brightnessBarFix:(NSInteger)state{
    state = (int)_brightness.value/6,25;
    switch ((int)state) {
        case 0:
            _image1.hidden=YES;
            _image2.hidden=YES;
            _image3.hidden=YES;
            _image4.hidden=YES;
            _image5.hidden=YES;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;

        case 1:
            _image1.hidden=NO;
            _image2.hidden=YES;
            _image3.hidden=YES;
            _image4.hidden=YES;
            _image5.hidden=YES;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 2:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=YES;
            _image4.hidden=YES;
            _image5.hidden=YES;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 3:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=YES;
            _image5.hidden=YES;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 4:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=YES;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 5:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=YES;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 6:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=YES;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 7:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=YES;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 8:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=YES;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 9:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=YES;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 10:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=YES;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 11:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=YES;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 12:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=NO;
            _image13.hidden=YES;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 13:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=NO;
            _image13.hidden=NO;
            _image14.hidden=YES;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 14:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=NO;
            _image13.hidden=NO;
            _image14.hidden=NO;
            _image15.hidden=YES;
            _image16.hidden=YES;
            break;
        case 15:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=NO;
            _image13.hidden=NO;
            _image14.hidden=NO;
            _image15.hidden=NO;
            _image16.hidden=YES;
            break;
        case 16:
            _image1.hidden=NO;
            _image2.hidden=NO;
            _image3.hidden=NO;
            _image4.hidden=NO;
            _image5.hidden=NO;
            _image6.hidden=NO;
            _image7.hidden=NO;
            _image8.hidden=NO;
            _image9.hidden=NO;
            _image10.hidden=NO;
            _image11.hidden=NO;
            _image12.hidden=NO;
            _image13.hidden=NO;
            _image14.hidden=NO;
            _image15.hidden=NO;
            _image16.hidden=NO;
            break;
    }
    
}
#pragma mark Media Keys
- (void)mediaKeyDown:(NSNumber *)mediaKey {
    switch (mediaKey.intValue) {
           
        case 120:
            if (_brightness){
                _brightness.value = MIN(_brightness.max, _brightness.value + _volume.max / 16);
                NSLog(@"Incresing brightness");
                [self showWindow];
                [self brightnessBarFix:_brightness.value];
            }
            break;
        case 122:
            if (_brightness){
                _brightness.value = MAX(0, _brightness.value - _brightness.max / 16);
                NSLog(@"Decresing brightness");
                [self showWindow];
                [self brightnessBarFix:_brightness.value];
            }
            break;
        case NX_KEYTYPE_MUTE:
            if (_mute)
            [self hideWindow];
            break;
        case NX_KEYTYPE_SOUND_UP:
            [self hideWindow];
            break;
        case NX_KEYTYPE_SOUND_DOWN:
           [self hideWindow];
            break;
    }
}
- (void)showWindow {
     NSWindow *mainWindow = [[[NSApplication sharedApplication] windows] objectAtIndex:0];
    [mainWindow setAlphaValue:0.9];
      if ([_timer isValid]){
            [_timer invalidate];
       }
      _timer = [NSTimer scheduledTimerWithTimeInterval:2 target:self selector:@selector(hideWindow) userInfo:nil repeats:NO];
    
}
- (void)hideWindow {
    NSWindow *mainWindow = [[[NSApplication sharedApplication] windows] objectAtIndex:0];
    [mainWindow setAlphaValue:0.0];
}

#pragma mark Observation
- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context{
    CGDirectDisplayID display = (CGDirectDisplayID)[_displayIDs pointerAtIndex:_displayNumber - 1];
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
        struct EDID edid = {};
        if (EDIDTest(display, &edid))
            dispatch_async(dispatch_get_main_queue(), ^{
                for (NSValue *value in @[[NSValue valueWithPointer:&edid.descriptor1], [NSValue valueWithPointer:&edid.descriptor2], [NSValue valueWithPointer:&edid.descriptor3], [NSValue valueWithPointer:&edid.descriptor4]]) {
                    union descriptor *des = value.pointerValue;
                    switch (des->text.type) {
                        case 0xFF:
                            assignWithNotice(self, serial, [DDC EDIDString:des->text.data]);
                            break;
                        case 0xFC:
                            assignWithNotice(self, name, [DDC EDIDString:des->text.data]);
                            break;
                    }
                }
                self.brightness = [[DDC alloc] initWithControl:BRIGHTNESS on:display];
                self.contrast = [[DDC alloc] initWithControl:CONTRAST on:display];
                self.volume = [[DDC alloc] initWithControl:AUDIO_SPEAKER_VOLUME on:display];
                self.mute = [[DDC alloc] initWithControl:AUDIO_MUTE on:display];
                self.reset = [[DDC alloc] initWithControl:RESET on:display];
                [self brightnessBarFix:_brightness.value];
            });
        else
            dispatch_async(dispatch_get_main_queue(), ^{
                assignWithNotice(self, serial, nil);
                assignWithNotice(self, name, nil);
                self.brightness = self.contrast = self.volume = self.mute = nil;
                NSLog(@"EDID Not Found Could not fetch the EDID, controls will be unavailable");
                
            });
    });
}

@end
